(* robot.ml *)

open Unix

(********************)
(* Type declaration *)
(********************)

type tile = Space | Water | Wall | Home
type base = {pos:int*int;mutable filled:bool}
type mode = GoHome | PickDeliver| Deliver | Nothing
type package = {idp:int;dnx:int;dny:int;wht:int}
type state = {mid:int;mutable cap:int;mutable pnx:int;mutable pny:int;
              mutable mon:int;mutable lpk:package list;
              mutable mde:mode;mutable lpi:(int*int)->int;
              mutable acc:(int*int)->bool}
type robot = {idn:int;mutable psx:int;mutable psy:int}
type action = Move of int | Pick of int | Drop of int | Set of int*int | Nop

(**********)
(* Parser *)
(**********)

let parse_tile = function
  | '.' -> (Space,false)
  | '#' -> (Wall,true)
  | '~' -> (Water,true)
  | '@' -> (Home,false)
  | _ -> assert false

let parse_row s =
  let l = ref [] in
  begin
    for i=0 to ((String.length s)-1) do
      l := !l@[parse_tile (String.get s i)]
    done;
    !l
  end

let gram = Grammar.create (Plexer.make ())
let dim = Grammar.Entry.create gram "dim"
let til = Grammar.Entry.create gram "til"
let row = Grammar.Entry.create gram "row"
let mrt = Grammar.Entry.create gram "mrt"
let pos = Grammar.Entry.create gram "pos"
let cnf = Grammar.Entry.create gram "cnf"
let pak = Grammar.Entry.create gram "pak"
let alp = Grammar.Entry.create gram "alp"
let act = Grammar.Entry.create gram "act"
let oas = Grammar.Entry.create gram "oas"
let ans = Grammar.Entry.create gram "ans"
EXTEND
  dim: [ [ col = INT; row = INT -> (int_of_string col,int_of_string row) ] ];
  til: [ [ "." -> (Space,false)
         | "~" -> (Water,true)
         | "#" -> (Wall,true)
         | "@" -> (Home,false) ] ];
  row: [ [ l = LIST1 til -> l ] ];
  mrt: [ [ ein = INT; ecp=INT; emn=INT ->
           {mid=int_of_string ein;
            cap=int_of_string ecp;
            pnx=0;
            pny=0;
            mon=int_of_string emn;
            lpk=[];
            mde=Nothing;
            lpi=(fun (x,y) -> 0);
            acc=fun c -> true} ] ];
  pos: [ [ "."; ein=INT; UIDENT "X"; epx=INT; UIDENT "Y"; epy=INT ->
           {idn=int_of_string ein;
            psx=(int_of_string epx)-1;
            psy=(int_of_string epy)-1} ] ];
  cnf: [ [ l = LIST1 pos -> l ] ];
  pak: [ [ eip=INT; edx=INT; edy=INT; ewt=INT ->
           {idp=int_of_string eip;
            dnx=(int_of_string edx)-1;
            dny=(int_of_string edy)-1;
            wht=int_of_string ewt} ] ];
  alp: [ [ l = LIST0 pak -> l ] ];
  act: [ [ UIDENT "N" -> Move 0
         | UIDENT "S" -> Move 1
         | UIDENT "E" -> Move 2
         | UIDENT "W" -> Move 3
         | UIDENT "P"; idn=INT -> Pick (int_of_string idn)
         | UIDENT "D"; idn=INT -> Drop (int_of_string idn)
         | UIDENT "X"; stx=INT; UIDENT "Y"; sty=INT ->
           Set ((int_of_string stx)-1,(int_of_string sty)-1) ] ];
  oas: [ [ "."; idn=INT; l = LIST0 act ->
           (int_of_string idn,if l=[] then [Nop] else l) ] ];
  ans: [ [ l = LIST1 oas -> l ] ];
END

let parse_dim s = Grammar.Entry.parse dim (Stream.of_string s)
(*let parse_row s = Grammar.Entry.parse row (Stream.of_string s)*)
let parse_mrt s = Grammar.Entry.parse mrt (Stream.of_string s)
let parse_cnf s = Grammar.Entry.parse cnf (Stream.of_string s)
let parse_alp s = Grammar.Entry.parse alp (Stream.of_string s)
let parse_ans s = Grammar.Entry.parse ans (Stream.of_string s)

let sub_sharp s = 
  try
  begin
    while true do
      String.set s (String.index s '#') '.'
    done;
    s
  end
  with Not_found -> s

(*************)
(* Game play *)
(*************)

let ask_package ic = parse_alp (input_line ic)

let ask_answer ic = 
  try 
    let t = input_line ic in
    begin
      (*print_endline ("Turn result: "^t);
      flush Pervasives.stdout;*)
      parse_ans (sub_sharp t)
    end
  with End_of_file ->
    begin
      print_endline "Game over";
      flush Pervasives.stdout;
      exit 0
    end

(**** Updates ****)

let rec find_my_state id = function
  | [] -> assert false
  | (idn,lac)::tl ->
    if id=idn then tl
    else (idn,lac)::(find_my_state id tl)

let update_position x y = function
  | 0 -> (x,y+1)
  | 1 -> (x,y-1)
  | 2 -> (x+1,y)
  | 3 -> (x-1,y)
  | _ -> assert false

let rec give_package id = function
  | [] -> assert false
  | e::tl ->
    if id = e.idp then e
    else give_package id tl

let rec extract_package id = function
  | [] -> assert false
  | e::tl ->
    if id = e.idp then tl
    else e::(extract_package id tl)

let update_mrt_act lav sta = function
  | Move n ->
    let (x,y)=update_position sta.pnx sta.pny n in
    begin
      sta.pnx <- x;
      sta.pny <- y
    end
  | Pick n ->
    let pkg = give_package n lav in
    begin
      sta.lpk <- sta.lpk@[pkg];
      sta.cap <- sta.cap - pkg.wht;
      sta.mde <- PickDeliver
    end
  | Drop n ->
    begin
      sta.cap <- sta.cap + (give_package n sta.lpk).wht;
      sta.lpk <- extract_package n sta.lpk
    end
  | Set(x,y) -> sta.pnx <- x;sta.pny <- y
  | Nop -> ()

let update_mrt lav sta lac = List.iter (update_mrt_act lav sta) lac

let update_rob_act rob = function
  | Move n -> let (x,y)=update_position rob.psx rob.psy n in
              rob.psx <- x;rob.psy <- y
  | Pick n -> ()
  | Drop n -> ()
  | Set(x,y) -> rob.psx <- x;rob.psy <- y
  | Nop -> ()

let update_rob rob lac = List.iter (update_rob_act rob) lac

let rec update_all_rob ans = function
  | [] -> []
  | e::tl ->
    (try update_rob e (List.assoc e.idn ans);e::(update_all_rob ans tl)
     with Not_found -> update_all_rob ans tl)

let rec mem_id_rob id = function
  | [] -> false
  | e::tl ->
    if e.idn = id then true
    else mem_id_rob id tl

let rec find_new_robots lrb = function
  | [] -> lrb
  | (id,lac)::tl ->
    if not(mem_id_rob id lrb) then
      let nrb = {idn=id;psx=0;psy=0} in
      begin
        update_rob nrb lac;
        find_new_robots (lrb@[nrb]) tl
      end
    else find_new_robots lrb tl

let main_update_rob lrb ans = find_new_robots (update_all_rob ans lrb) ans

let update sta rob lav ans =
  begin
    update_mrt lav sta (List.assoc sta.mid ans);
    let nlb = main_update_rob rob (find_my_state sta.mid ans) in
    (sta,nlb)
  end

(**** Our tactic ****)

let move = function
  | 0 -> "Move N"
  | 1 -> "Move S"
  | 2 -> "Move E"
  | 3 -> "Move W"
  | _ -> assert false

let pack act =
  let rec pack_rec = function
  | [] -> ""
  | id::tl -> " "^(string_of_int id)^(pack_rec tl) in
  fun lpk -> act^(pack_rec lpk)

let action bid act sta oc =
  begin
    sta.mon <- sta.mon - bid;
    output_string oc ((string_of_int bid)^" "^act^"\n");
    flush oc
  end

(******************************************)
(* Naive algorithm for lightning division *)
(******************************************)

let good_destination (x,y) pkg =
  if (x=pkg.dnx) && (y=pkg.dny) then pkg.idp
  else failwith "Not yet"

let try_drop (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  let l = List.fold_left
          (fun a e -> try a@[good_destination (sta.pnx,sta.pny) e]
                      with _ -> a) [] sta.lpk in
  if l = [] then failwith "Nothing to drop"
  else action 1 (pack "Drop" l) sta oc
  
let try_pick (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  let rec lpk_rec cap = function
    | [] -> []
    | pkg::tl ->
      if pkg.wht<=cap && (sta.acc (pkg.dnx,pkg.dny)) then
        pkg.idp::(lpk_rec (cap-pkg.wht) tl)
      else lpk_rec cap tl in
  let l = lpk_rec sta.cap lav in
  if l = [] then failwith "Nothing to pick"
  else action 1 (pack "Pick" l) sta oc

let ok_case (xdm,ydm) tab (x,y) =
  not begin
    x<0 || x>=xdm || y<0 || y >=ydm ||
    let typ = fst (tab.(x).(y)) in
    typ=Water || typ=Wall
  end

let verify_dir (xdm,ydm) tab (x,y) = function
  | 0 -> ok_case (xdm,ydm) tab (x,y+1)
  | 1 -> ok_case (xdm,ydm) tab (x,y-1)
  | 2 -> ok_case (xdm,ydm) tab (x+1,y)
  | 3 -> ok_case (xdm,ydm) tab (x-1,y)
  | _ -> assert false

let rec try_move (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  let dir = Random.int 4 in
  if verify_dir (xdm,ydm) tab (sta.pnx,sta.pny) dir then
    action 1 (move dir) sta oc
  else try_move (xdm,ydm,tab,blt,sta,rob,gpi) lav oc

let drinker_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  try try_drop (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with _ ->
  try try_pick (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with _ ->
  try_move (xdm,ydm,tab,blt,sta,rob,gpi) lav oc

(*********************************)
(* Algorithm to deliver packages *)
(*********************************)

let print_tile = function
 | Water -> "~"
 | Wall -> "#"
 | Space -> "."
 | Home -> "@"

let print_tab (xdm,ydm) tab = 
  for j=ydm-1 downto 0 do
    for i=0 to xdm-1 do
      print_string (print_tile (fst (tab.(i).(j))));
      flush Pervasives.stdout;
    done;
    Printf.printf "\n";
    flush Pervasives.stdout;
  done

(**** Rectangulation of the board ****)

let test_xline map xmin xmax y =
  let x = ref xmin in
  while !x < xmax && not map.(!x).(y) do incr x done ;
  !x = xmax ;;

let test_yline map x ymin ymax =
  let y = ref ymin in
  while !y < ymax && not map.(x).(!y) do incr y done ;
  !y = ymax ;;

let find_optimal_square map x0 y0 lim =
  let n = ref 1 in
  while !n < lim &&
    test_xline map x0 (x0 + !n + 1) (y0 + !n) &&
    test_yline map (x0 + !n) y0 (y0 + !n)
  do incr n done ; !n ;;

let find_optimal_xrect map x0 y0 xdm ht =
  let x = ref (x0 + ht) in
  while !x < xdm && test_yline map !x y0 (y0 + ht)
  do incr x done ; (!x - x0) ;;

let find_optimal_yrect map x0 y0 wd ydm =
  let y = ref (y0 + wd) in
  while !y < ydm && test_xline map x0 (x0 + wd) !y
  do incr y done ; (!y - y0) ;;

let find_optimal_rect map x0 y0 xdm ydm =
  assert (xdm > 0 && ydm > 0) ;
  assert (x0 >= 0 && x0 < xdm && y0 >= 0 && y0 < ydm) ;
  assert (not map.(x0).(y0)) ;
  let n = find_optimal_square map x0 y0 (min (xdm - x0) (ydm - y0)) in
  let n1 = find_optimal_xrect map x0 y0 xdm n in
  let n2 = find_optimal_yrect map x0 y0 n ydm in
  if n1 >= n2
  then (x0, x0 + n1, y0, y0 + n)
  else (x0, x0 + n, y0, y0 + n2) ;;

let fill_rect map (xmin, xmax, ymin, ymax) =
  for y = ymin to ymax - 1 do
    for x = xmin to xmax - 1 do
      map.(x).(y) <- true
    done
  done ;;

let rec rev_number accu n = function
  | [] -> accu
  | x :: tl -> rev_number ((n, x) :: accu) (n - 1) tl ;;

let rectangulation (xdm, ydm) tab =
  if xdm = 0 || ydm = 0 then [] else begin
    (* building the temp map *)
    let map = Array.make_matrix xdm ydm false in
    for y=0 to ydm - 1 do
      for x=0 to xdm - 1 do
        match fst tab.(x).(y) with
        | Space | Home -> ()
        | Wall  | Water -> map.(x).(y) <- true
      done
    done ;
    let stack = ref [] in
    let count = ref 0 in
    let x = ref 0 in
    let y = ref 0 in
    while !y < ydm do
      (* look for the next free pixel *)
      while !y < ydm && map.(!x).(!y) do
	incr x ;
	if !x = xdm then begin x := 0 ; incr y end
      done ;
      if !y < ydm then begin
	let rect = find_optimal_rect map !x !y xdm ydm in
	fill_rect map rect ;
	let (xmin, xmax, ymin, ymax) = rect in
	stack := rect :: !stack ; incr count ;
	incr x ;
	if !x = xdm then begin x := 0 ; incr y end
      end
    done ;
    rev_number [] (!count - 1) !stack
  end ;;

let rec which_rectangle (i,j) = function
  | [] -> raise Not_found
  | (id,(xmin,xmax,ymin,ymax))::tl ->
    if (i>=xmin && i<xmax && j>=ymin && j<ymax) then id
    else which_rectangle (i,j) tl

let rec print_rectangle = function
  | [] -> ()
  | (id,(xmin,xmax,ymin,ymax)):: tl ->
    begin
      Printf.printf "Rectangle No %i: (%i,%i,%i,%i)\n" id xmin xmax ymin ymax;
      flush Pervasives.stdout;
      print_rectangle tl
    end

(**** Computation of the gates ****)

exception NoGate

let maxi (a,b) = if a>=b then a else b
let mini (a,b) = if a<=b then a else b

let lgt = ref []

let fill_tables lrc tabx taby =
  List.iter
  (fun (id,(xmin,xmax,ymin,ymax)) ->
   begin
     Hashtbl.add tabx xmin id;
     Hashtbl.add tabx xmax id;
     Hashtbl.add taby ymin id;
     Hashtbl.add taby ymax id
   end) lrc

let verify_add tab id ele =
  if not(List.mem ele (Hashtbl.find_all tab id)) then
    Hashtbl.add tab id ele

let gate_x lrc id0 id1 xk tabg =
  let (xmin0,xmax0,ymin0,ymax0) = List.assoc id0 lrc
  and (xmin1,xmax1,ymin1,ymax1) = List.assoc id1 lrc in
  let ymin=maxi(ymin0,ymin1)
  and ymax=mini(ymax0,ymax1) in
  if ymin<ymax then
    let gat = (ymin+ymax)/2 in
    begin
      if xmin0<xmin1 then
      begin      
        verify_add tabg id0 (xk-1,gat);
        verify_add tabg id1 (xk,gat)
      end
      else
      begin
        verify_add tabg id0 (xk,gat);
        verify_add tabg id1 (xk-1,gat)
      end;
      lgt := ((xk,gat),(xk-1,gat)) :: !lgt
    end

let gate_y lrc id0 id1 yk tabg =
  let (xmin0,xmax0,ymin0,ymax0) = List.assoc id0 lrc
  and (xmin1,xmax1,ymin1,ymax1) = List.assoc id1 lrc in
  let xmin=maxi(xmin0,xmin1)
  and xmax=mini(xmax0,xmax1) in
  if xmin<xmax then
    let gat = (xmin+xmax)/2 in
    begin
      if ymin0<ymin1 then
      begin      
        verify_add tabg id0 (gat,yk-1);
        verify_add tabg id1 (gat,yk)
      end
      else
      begin
        verify_add tabg id0 (gat,yk);
        verify_add tabg id1 (gat,yk-1)
      end;
      lgt := ((gat,yk),(gat,yk-1)) :: !lgt
    end

let gate_all_x lrc lid xk tabg =
  let rec gate_rec = function
    | [] -> ()
    | e::tl ->
      begin
        List.iter (fun e' -> if not(e=e') then gate_x lrc e e' xk tabg) lid;
        gate_rec tl
      end in
  gate_rec lid

let gate_all_y lrc lid yk tabg =
  let rec gate_rec = function
    | [] -> ()
    | e::tl ->
      begin
        List.iter (fun e' -> if not(e=e') then gate_y lrc e e' yk tabg) lid;
        gate_rec tl
      end in
  gate_rec lid

let compute_gate_x lrc tabx tabg =
  Hashtbl.iter
    (fun k e ->
       let lid = Hashtbl.find_all tabx k in
       gate_all_x lrc lid k tabg) tabx

let compute_gate_y lrc taby tabg =
  Hashtbl.iter
    (fun k e ->
       let lid = Hashtbl.find_all taby k in
       gate_all_y lrc lid k tabg) taby

let compute_gate lrc tabx taby tabg =
  begin
    compute_gate_x lrc tabx tabg;
    compute_gate_y lrc taby tabg
  end

let rec give_list tabg = function
  | [] -> []
  | (id,rct)::tl -> (id,Hashtbl.find_all tabg id)::(give_list tabg tl)

let gate_list lrc =
  let tabx = Hashtbl.create 53
  and taby = Hashtbl.create 53
  and tabg = Hashtbl.create 53 in
  begin
    fill_tables lrc tabx taby;
    compute_gate lrc tabx taby tabg;
    (give_list tabg lrc,!lgt)
  end

let print_gates =
  List.iter
  (fun (id,l) ->
     begin
       Printf.printf "Rectangle No %i\n" id;
       flush Pervasives.stdout;
       List.iter
       (fun (x,y) ->
          begin
            Printf.printf "Porte (%i,%i)\n" x y;
            flush Pervasives.stdout
          end) l
     end)

(**** Construction of the gate graph ****)

let infinity = 1000001

(*** Types ***)

type node = {
    n_id : int ;
    n_pos : int * int ;
    mutable n_succ : (int * node) list
  }

let dummy_node =
  { n_id = -1 ; n_pos = (-1, -1) ; n_succ = [] }

(*** Distance ***)

let distance (x0, y0) (x1, y1) =
  let dx = x1 - x0 in
  let dy = y1 - y0 in
  (if dx >= 0 then dx else -dx) +
  (if dy >= 0 then dy else -dy)

let make_graph lrc pairs =
  let num_rectangles = List.length lrc
  and htable = Hashtbl.create 257 in
  let cur_id = ref 0 in
  let register_pos pos =
    if not (Hashtbl.mem htable pos) then begin
      Hashtbl.add htable pos !cur_id ;
      incr cur_id
    end in
  (* register all positions *)
  List.iter begin
    function (pos0, pos1) ->
      register_pos pos0 ;
      register_pos pos1
  end pairs ;
  (* build the node table (we now know its size) *)
  let num_nodes = !cur_id in
  let node_table = Array.make num_nodes dummy_node in
  Hashtbl.iter begin
    fun pos id ->
      node_table.(id) <- { n_id = id ; n_pos = pos ; n_succ = [] }
  end htable ;
  (* connect adjacent gates *)
  List.iter begin
    function (pos0, pos1) ->
      let nod0 = node_table.(Hashtbl.find htable pos0) in
      let nod1 = node_table.(Hashtbl.find htable pos1) in
      nod0.n_succ <- (1, nod1) :: nod0.n_succ ;
      nod1.n_succ <- (1, nod0) :: nod1.n_succ
  end pairs ;
  (* building the gate table *)
  let gate_table = Array.make num_rectangles [] in
  for id = 0 to num_nodes - 1 do
    let rect_id = which_rectangle node_table.(id).n_pos lrc in
    gate_table.(rect_id) <- id :: gate_table.(rect_id)
  done ;
  (* filling the rest of the graph *)
  for i = 0 to num_rectangles - 1 do
    let gates = gate_table.(i) in
    List.iter begin
      function id0 ->
        List.iter begin
          function id1 ->
            if id1 <> id0 then begin
              let nod0 = node_table.(id0) in
              let nod1 = node_table.(id1) in
              let dist = distance nod0.n_pos nod1.n_pos in
              nod0.n_succ <- (dist, nod1) :: nod0.n_succ
              (* symmetric case will be handled in another
                 iteration of the same double loop *)
            end
        end gates
    end gates
  done ;
  (node_table, gate_table)

(*** Compute the distance table ***)

let make_dist_table lrc num_nodes (node_table,gate_table) finish =
  (* compute the starting values *)
  let rect_id =
    try which_rectangle finish lrc
    with Not_found -> assert false
  in
  let init_list = gate_table.(rect_id) in
  let table = Array.make num_nodes infinity in
  List.iter begin
    function id ->
      let pos = node_table.(id).n_pos in
      table.(id) <- distance pos finish
  end init_list ;
  (* compute the transitive closure *)
  let flag = ref true in
  while !flag do
    flag := false ;
    for id = 0 to num_nodes - 1 do
      if not (List.mem id init_list) then begin
        let min_v = ref infinity in
        (* look at all the neighbors... *)
        List.iter begin
          function (k, node) ->
            let v = k + table.(node.n_id) in
            if v < !min_v then min_v := v
        end node_table.(id).n_succ ;
        (* ... and record the min *)
        if table.(id) <> !min_v then begin
          table.(id) <- !min_v ;
          flag := true
        end
      end
    done
  done ;
  table

exception NoPath

let make_phi lrc num_nodes (node_table,gate_table) finish=
  let dist_table=make_dist_table lrc num_nodes (node_table,gate_table)
                 finish in
  let rect_fin =
    try which_rectangle finish lrc
    with Not_found -> assert false
  in
  function pos ->
    let rect_id=which_rectangle pos lrc in
    if rect_id=rect_fin then
      distance pos finish
    else 
      let d_min=ref infinity in
      begin
        List.iter
          (function id ->
            let d=(distance pos node_table.(id).n_pos)+dist_table.(id) in
            if d<(!d_min) then d_min:=d) gate_table.(rect_id);
        !d_min
      end

let generic_phi lrc pairs =
  let (node_table,gate_table)=make_graph lrc pairs in
  make_phi lrc (Array.length node_table) (node_table,gate_table)

(*** Accessibility ***)

(* make_access_table : int * int -> bool array *)
let make_access_table lrc (node_table,gate_table) start =
  let num_rectangles = List.length lrc
  and num_nodes = Array.length node_table in
  let v = Array.make num_nodes false in
  (* start rectangle *)
  let (x, y) = start in
  let start_id = which_rectangle start lrc in
  (* mark as accessible all the gates here *)
  List.iter begin
    fun id ->
      v.(id) <- true
  end gate_table.(start_id) ;
  (* perform the transitive closure *)
  let modified = ref true in
  while !modified do
    modified := false ;
    for id = 0 to num_nodes - 1 do
      if not v.(id) then begin
	List.iter begin
	  function (_, node) ->
	    if v.(node.n_id) then begin
	      v.(id) <- true ;
	      modified := true
	    end
	end node_table.(id).n_succ
      end
    done
  done ;
  (* building the table *)
  let table = Array.make num_rectangles false in
  table.(start_id) <- true ; (* start rectangle *)
  for rect_id = 0 to num_rectangles - 1 do
    match gate_table.(rect_id) with
    | [] -> ()
    | id :: _ -> if v.(id) then table.(rect_id) <- true
  done ;
  table ;;

(* make_access_fun : int * int -> (int * int -> bool) *)
let make_access_fun lrc (node_table,gate_table) start =
  let access_table = make_access_table lrc (node_table,gate_table) start in
  function pos ->
    try
      let rect_id = which_rectangle pos lrc in
      access_table.(rect_id)
    with Not_found -> false ;;

let graph_prim lrc pairs start =
  let (node_table,gate_table)=make_graph lrc pairs in
  (make_phi lrc (Array.length node_table) (node_table,gate_table),
   make_access_fun lrc (node_table,gate_table) start)

(**********************)
(* The gate algorithm *)
(**********************)

let default_algo = drinker_algo

let verify_robot rob (x,y) = List.exists (fun r -> r.psx=x && r.psy=y) rob

let move = function
  | 0 -> "Move N"
  | 1 -> "Move S"
  | 2 -> "Move E"
  | 3 -> "Move W"
  | _ -> assert false

let bid_defense sta = sta.mon/100

let escape_x (xdm,ydm) tab (x,y) dir sta oc =
  if ok_case (xdm,ydm) tab (x+1,y) then
    action (bid_defense sta) (move 2) sta oc
  else if ok_case (xdm,ydm) tab (x-1,y) then
    action (bid_defense sta) (move 3) sta oc
  else
    action (bid_defense sta) (move dir) sta oc

let escape_y (xdm,ydm) tab (x,y) dir sta oc =
  if ok_case (xdm,ydm) tab (x,y+1) then
    action (bid_defense sta) (move 0) sta oc
  else if ok_case (xdm,ydm) tab (x,y-1) then
    action (bid_defense sta) (move 1) sta oc
  else
    action (bid_defense sta) (move dir) sta oc

let verify_tile (xdm,ydm) tab sta rob (x,y) oc =
  try
     let til0=fst (tab.(x-1).(y))
     and til1=fst (tab.(x+1).(y)) in
     if (til0=Water && (verify_robot rob (x+1,y))) then
       escape_y (xdm,ydm) tab (x,y) 2 sta oc
     else if ((verify_robot rob (x-1,y)) && til1=Water) then
       escape_y (xdm,ydm) tab (x,y) 3 sta oc
     else failwith "No danger"
  with _ ->
  (try
     let til0=fst (tab.(x).(y-1))
     and til1=fst (tab.(x).(y+1)) in
     if (til0=Water && (verify_robot rob (x,y+1))) then
       escape_y (xdm,ydm) tab (x,y) 0 sta oc
     else if ((verify_robot rob (x,y-1)) && til1=Water) then
       escape_y (xdm,ydm) tab (x,y) 1 sta oc
     else failwith "No danger"
   with _ -> failwith "No danger")

let try_defense (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  try verify_tile (xdm,ydm) tab sta rob (sta.pnx,sta.pny) oc
  with Failure _ -> failwith "No defense needed"

let choose_dir ldir =
  let bnd = List.length ldir in
  fst (List.nth ldir (Random.int bnd))

let rec build_min_list min lacc = function
  | [] -> lacc
  | (i,min')::tl ->
    if min=min' then build_min_list min (lacc@[(i,min')]) tl
    else build_min_list min lacc tl

let min_list ldir =
  let rec min_rec min = function
    | [] -> min
    | (_,min')::tl ->
      if min'<min then min_rec min' tl
      else min_rec min tl in
  let min = min_rec (snd (List.hd ldir)) ldir in
  choose_dir (build_min_list min [] ldir)

let which_direction (xdm,ydm) tab sta =
  let ldir =
  (if verify_dir (xdm,ydm) tab (sta.pnx,sta.pny) 0 then
   [(0,sta.lpi (sta.pnx,sta.pny+1))] else [])@
  (if verify_dir (xdm,ydm) tab (sta.pnx,sta.pny) 1 then
   [(1,sta.lpi (sta.pnx,sta.pny-1))] else [])@
  (if verify_dir (xdm,ydm) tab (sta.pnx,sta.pny) 2 then
   [(2,sta.lpi (sta.pnx+1,sta.pny))] else [])@
  (if verify_dir (xdm,ydm) tab (sta.pnx,sta.pny) 3 then
   [(3,sta.lpi (sta.pnx-1,sta.pny))] else []) in
  move (min_list ldir)

let rec close_base (xb,yb) = function
  | [] -> assert false
  | a::tl ->
    if a.pos=(xb,yb) then a.filled <-false
    else close_base (xb,yb) tl

let rec next_base sta = function
  | [] -> failwith "No base"
  | a::tl ->
    if a.filled && (sta.acc a.pos) then a.pos
    else next_base sta tl

let home_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  if (sta.lpi (sta.pnx,sta.pny))=0 then
  begin
    try try_pick (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with Failure _ ->
    try 
      begin
        close_base (sta.pnx,sta.pny) blt;
        let (xb,yb) = next_base sta blt in
        begin
          sta.lpi <- gpi (xb,yb);
          action 1 (which_direction (xdm,ydm) tab sta) sta oc
        end
      end
    with Failure _ ->
    begin
      sta.mde <- Nothing;
      default_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc
    end
  end
  else action 1 (which_direction (xdm,ydm) tab sta) sta oc

let next_package sta gpi =
  if sta.lpk=[] then
    failwith "No more package"
  else
    let pkg=List.hd sta.lpk in
    sta.lpi <- gpi (pkg.dnx,pkg.dny)

let rec deliver_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  if (sta.lpi (sta.pnx,sta.pny))=0 then
    try try_drop (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with Failure _ ->
    (try (next_package sta gpi);
         deliver_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with Failure _ ->
    (try 
      let (xb,yb) = next_base sta blt in
      begin
        sta.lpi <- gpi (xb,yb);
        home_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc
      end with Failure _ -> default_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc))
  else action 1 (which_direction (xdm,ydm) tab sta) sta oc

let pick_deliver_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  let rec pkg_path = function
    | [] -> raise NoPath
    | pkg::tl ->
      (try sta.lpi <- gpi (pkg.dnx,pkg.dny)
       with NoPath -> pkg_path tl) in
  begin
    pkg_path sta.lpk;
    sta.mde <- Deliver;
    deliver_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc
  end

let gate_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  try try_defense (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with Failure _ ->
  (try try_pick (xdm,ydm,tab,blt,sta,rob,gpi) lav oc with Failure _ ->
  (match sta.mde with
   | GoHome -> home_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc
   | PickDeliver -> pick_deliver_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc
   | Deliver -> deliver_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc
   | Nothing -> default_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc))

(**************)
(* The tactic *)
(**************)

let tactic (xdm,ydm,tab,blt,sta,rob,gpi) lav oc =
  gate_algo (xdm,ydm,tab,blt,sta,rob,gpi) lav oc

(*************)
(* Main loop *)
(*************)

let cpu_warning act oc =
  let t0 = Sys.time () in
  act oc;
  let t1 = Sys.time () in
  if (t1-.t0)>1. then
  begin
    print_endline "Warning: more than 1-CPUs for this turn";
    flush Pervasives.stdout
  end

let print_location sta =
  begin
    Printf.printf "Location: (%i,%i)\n" sta.pnx sta.pny;
    flush Pervasives.stdout
  end

let print_package lav =
  begin
    print_string "Package:";
    if lav=[] then print_string " none"
    else List.iter (fun pkg -> print_string " ";print_int pkg.idp) lav;
    print_newline ();
    flush Pervasives.stdout
  end

let rec main_loop (xdm,ydm,tab,blt,sta,rob,gpi) ic oc =
  begin
    (*print_location sta;*)
    let lav = ask_package ic in
    (*print_package lav;*)
    cpu_warning (tactic (xdm,ydm,tab,blt,sta,rob,gpi) lav) oc;
    let ans = ask_answer ic in
    let (nsta,nrob) = update sta rob lav ans in
    main_loop (xdm,ydm,tab,blt,nsta,nrob,gpi) ic oc
  end

(******************)
(* Initialization *)
(******************)

let rec init acc sta = function
  | [] -> assert false
  | e::tl ->
    if sta.mid = e.idn then
    begin
      sta.pnx <- e.psx;
      sta.pny <- e.psy;
      (sta,acc@tl)
    end
    else init (acc@[e]) sta tl

let base_list (xdm,ydm) tab =
  let l = ref [] in
  begin
    for i=0 to (xdm-1) do
      for j=0 to (ydm-1) do
        if fst(tab.(i).(j))=Home then l := !l@[{pos=(i,j);filled=true}]
      done
    done;
    !l
  end

let read_board ic =
  let (xdm,ydm) = parse_dim (input_line ic) in
  let tab = Array.make_matrix xdm ydm (Space,false) in
  begin
    for i=0 to (ydm-1) do
      let l = parse_row (input_line ic) in
      for j=0 to (xdm-1) do
        tab.(j).(i) <- List.nth l j
      done
    done;
    let blt = base_list (xdm,ydm) tab
    and sta = parse_mrt (input_line ic) in
    let cnf = parse_cnf (sub_sharp (input_line ic)) in
    let (nsta,ncnf) = init [] sta cnf in
    (xdm,ydm,tab,blt,nsta,ncnf)
  end

let init_mode lbs sta (gpi,acc) =
  begin
    sta.acc <- acc;
    try
      let (xb,yb) = next_base sta lbs in
      begin
        sta.lpi <- gpi (xb,yb);
        sta.mde <- GoHome
      end
    with Failure _ -> sta.mde <- Nothing
  end

let initialization ic oc =
  begin
    output_string oc "Player\n";
    flush oc;
    let (xdm,ydm,tab,blt,nsta,ncnf) = read_board ic in
    (*print_tab (xdm,ydm) tab;*)
    let lrc = rectangulation (xdm,ydm) tab in
    (*print_rectangle lrc;*)
    let (lgt,lgc) = gate_list lrc in
    (*print_gates lgt;*)
    let (gpi,acc) = graph_prim lrc lgc (nsta.pnx, nsta.pny) in
    init_mode blt nsta (gpi,acc);
    (*print_endline "Initialization...Done";
    flush Pervasives.stdout;*)
    (xdm,ydm,tab,blt,nsta,ncnf,gpi)
  end

(****************)
(* Socket stuff *)
(****************)

let domain_of = function
  | ADDR_UNIX _ -> PF_UNIX
  | ADDR_INET _ -> PF_INET

let open_connection sockaddr =
  let domain = domain_of sockaddr in
  let sock = Unix.socket domain Unix.SOCK_STREAM 0 
  in try Unix.connect sock sockaddr ;
     (Unix.in_channel_of_descr sock , Unix.out_channel_of_descr sock)
     with exn -> Unix.close sock ; raise exn

let shutdown_connection inchan =
  Unix.shutdown (Unix.descr_of_in_channel inchan) Unix.SHUTDOWN_SEND ;;

let main_client client_fun =
  if Array.length Sys.argv < 3 
  then Printf.printf "Usage: runme serveur port\n"
  else let serveur = Sys.argv.(1) in
  let serveur_adr =
    try  Unix.inet_addr_of_string serveur 
    with Failure("inet_addr_of_string") -> 
      try  (Unix.gethostbyname serveur).Unix.h_addr_list.(0) 
      with Not_found ->
        Printf.eprintf "%s : serveur inconnu\n" serveur ;
        exit 2
  in try 
     let port = int_of_string (Sys.argv.(2)) in
     let sockadr = Unix.ADDR_INET(serveur_adr,port) in 
     let ic,oc = open_connection sockadr
     in client_fun ic oc ;
     shutdown_connection ic
     with Failure("int_of_string") -> Printf.eprintf "bad port number";
                                      exit 2
;;

let main ic oc = main_loop (initialization ic oc) ic oc

let _ =
  begin
    Random.self_init ();
    main_client main
  end
